import os
import platform
import subprocess
import sys
import time
import unittest

#
# Prerequisites:
# sudo apt-get -y install lxc libvirt-bin make gcc


class Lxc:
    def __init__(self, template="ubuntu", relnum="12.04", release=None,
                 config=None, name=None):
        self.name = name
        if not name:
            if release:
                self.name = "test-%s" % release
            else:
                self.name = "test-container"
        self.release = release
        self.config = config
        self.template = template

    # Clone - for now only supports copies of directory backed containers
    # add lvm and btrfs supports
    def Clone(self, clonefrom):
        if not clonefrom:
            return ValueError, "Clone requires clonefrom"

        cmd = ["lxc-clone", "-o", clonefrom.name, "-n", self.name]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            return ValueError, ("Failed cloning %s from %s" %
                                (self.name, clonefrom.name))

    def Create(self, config=None):
        cmd = ["lxc-create", "-t", self.template, "-n", self.name]
        if self.config:
            cmd.append("-f")
            cmd.append(self.config)
        if self.release:
            cmd.append("--")
            cmd.append("-r %s" % self.release)

        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed creating %s" % (self.name))

    def Destroy(self, ignorefail=False):
        ret = subprocess.call(["lxc-destroy", "-f", "-n", self.name],
                              stdout=None, stderr=None)
        if ret != 0 and not ignorefail:
            raise ValueError("Failed destroying %s" % (self.name))

    def PkgSetup(self):
        rootpath = "/var/lib/lxc/%s/rootfs" % (self.name)
        cmd = ["chroot", rootpath, "apt-get", "update"]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed apt-get update in container %s" %
                             (self.name))

        cmd = ["chroot", rootpath, "apt-get", "-y", "install", "libcap2"]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed installing libcap2 in container %s" %
                             (self.name))
        cmd = ["chroot", rootpath, "apt-get", "-y", "install",
               "--download-only", "lxc"]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed downloading lxc in container %s" %
                             (self.name))

        cmd = ["chroot", rootpath, "apt-get", "-y", "install", "libapparmor1"]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed installing libapparmor1 in container %s" %
                             (self.name))

        # extract lxc package into container's /tmp
        pkgpath = "%s/var/cache/apt/archives/lxc_*" % (rootpath)
        tmppath = "%s/tmp" % (rootpath)
        cmd = "dpkg -x %s %s" % (pkgpath, tmppath)
        ret = subprocess.call(cmd, shell=True, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed extracting lxc in container %s" %
                             (self.name))

        # copying /usr/lib/lxc into container
        srcpath = "%s/usr/lib/lxc" % (tmppath)
        destpath = "%s/usr/lib" % (rootpath)
        cmd = ["cp", "-a", srcpath, destpath]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed copying lxc package contents in"
                             "container %s" % (self.name))

    def Execute(self, program="/bin/true"):
        cmd = ["lxc-execute", "-n", self.name, "--", program]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed executing %s in %s" %
                             (program, self.name))

    def Start(self, init="/sbin/init"):
        cmd = ["lxc-start", "-d", "-n", self.name]
        if init != "/sbin/init":
            cmd.append("--")
            cmd.append(init)
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed starting %s in %s" % (init, self.name))

    def ConfirmRunning(self, notrunning=False):
        # ideally lxc-wait would take a timeout and we could do
        # lxc-wait -n test-$release -s RUNNING -t 5.  We don't though.
        time.sleep(2)

        cmd1 = ["lxc-info", "-n", self.name, "-s"]
        sp = subprocess.Popen(cmd1, stderr=subprocess.STDOUT,
                              stdout=subprocess.PIPE)
        output = sp.communicate()[0]
        if not notrunning:
            output.index("RUNNING")  # will raise a ValueError if not found.
        else:
            output.index("STOPPED")  # will raise a ValueError if not found.

    def TrySsh(self):
        True
        # not yet implemented

    # should we be trying a clean shutdown?
    def Stop(self):
        cmd = ["lxc-stop", "-n", self.name]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed stopping %s" % (self.name))

    def apparmor_profile(self, profile):
        # remove any existing lxc.aa_profile lines, and add a new one
        # saying lxc.aa_profile = unconfined.
        configfile = "/var/lib/lxc/%s/config" % (self.name)
        cmd = ["sed", "-i", "/aa_profile/d", configfile]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed to remove existing aa_profile lines")
        newprofile = '''$a\
lxc.aa_profile = unconfined'''
        cmd = ["sed", "-i", newprofile, configfile]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed to set unconfined profile")


class ApparmorTests(unittest.TestCase):
    def setUp(self):
        self.lxc = Lxc(config=None, release="precise")
        self.lxc.Destroy(ignorefail=True)
        self.lxc.Create()
        self.lxc.apparmor_profile("unconfined")

    def tearDown(self):
        self.lxc.Destroy(ignorefail=True)

    def runTest(self):
        try:
            self.lxc.Start()
            self.lxc.ConfirmRunning()
            self.lxc.Stop()
        except ValueError:
            #self.assertTrue(False, "Failed apparmor test with error %s" % (e)
            print "Apparmor test failed as expected (see LP bug 987371)"

if __name__ == '__main__':
    # set keepcache true if you're developing and want to save some time...
    keepcache = False
    # check for uid 0
    uid = os.getuid()
    if uid != 0:
        print >>sys.stderr, "Run this test suite as root"
        sys.exit(1)

    # worry about what to do about other distros later
    dist = platform.dist()
    if dist[0] != "Ubuntu":
        print >>sys.stderr, ("This test suite does not yet support %s" %
                             (dist[0]))
        sys.exit(1)

    # make sure the needed prereqs are installed
    cmd = ["sudo", "apt-get", "-y", "install", "lxc", "libvirt-bin", "make",
           "gcc"]
    if dist[1] >= 12.10:
        cmd += ["bzr", "python3-lxc"]

    ret = subprocess.call(cmd, stdout=None, stderr=None)
    if ret != 0:
        print "Failed installing prerequisites"
        sys.exit(1)

    # apt-get install -y cgroup-lite libvirt-bin lxc lsb-release make gcc
    if dist[1] < 11.10:
        print >>sys.stderr, "testsuite only supported on oneiric and precise"
        sys.exit(1)

    suite = unittest.TestSuite()

    if dist[1] >= 12.04:
        suite.addTest(ApparmorTests())

    rc = unittest.TextTestRunner(verbosity=2).run(suite)

    if not rc.wasSuccessful():
        sys.exit(1)

# set et sw=4 ts=4 ai si sm background=dark
