#!/usr/bin/python

import os
import platform
import subprocess
import sys
import time
import unittest

#
# Prerequisites:
# sudo apt-get -y install lxc libvirt-bin make gcc


class Lxc:
    def __init__(self, template="ubuntu", relnum="12.04", release=None,
                 config=None, name=None):
        self.name = name
        if not name:
            if release:
                self.name = "test-%s" % release
            else:
                self.name = "test-container"
        self.release = release
        self.config = config
        self.template = template

    # Clone - for now only supports copies of directory backed containers
    # add lvm and btrfs supports
    def Clone(self, clonefrom):
        if not clonefrom:
            return ValueError("Clone requires clonefrom")

        cmd = ["lxc-clone", "-o", clonefrom.name, "-n", self.name]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            return ValueError("Failed cloning %s from %s" %
                              (self.name, clonefrom.name))

    def Create(self, config=None):
        cmd = ["lxc-create", "-t", self.template, "-n", self.name]
        if self.config:
            cmd.append("-f")
            cmd.append(self.config)
        if self.release:
            cmd.append("--")
            cmd.append("-r %s" % self.release)

        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed creating %s" % (self.name))

    def Destroy(self, ignorefail=False):
        ret = subprocess.call(["lxc-destroy", "-f", "-n", self.name],
                              stdout=None, stderr=None)
        if ret != 0 and not ignorefail:
            raise ValueError("Failed destroying %s" % (self.name))

    def PkgSetup(self):
        rootpath = "/var/lib/lxc/%s/rootfs" % (self.name)
        cmd = ["chroot", rootpath, "apt-get", "update"]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed apt-get update in container %s" %
                             (self.name))

        override='''manual'''
        f=open("/var/lib/lxc/%s/rootfs/etc/init/lxc.override" % (self.name), "w")
        f.write(override)
        f.close()
        f=open("/var/lib/lxc/%s/rootfs/etc/init/lxc-net.override" % (self.name), "w")
        f.write(override)
        f.close()
        updatetxt='''#!/bin/sh
exit 101
'''
        f=open("/var/lib/lxc/%s/rootfs/usr/sbin/policy-rc.d" % (self.name), "w")
        f.write(updatetxt)
        f.close()
        cmd = ["chmod", "+x", "/var/lib/lxc/%s/rootfs/usr/sbin/policy-rc.d" % (self.name)]
        ret = subprocess.call(cmd, stdout=None, stderr=None)

        cmd = ["chroot", rootpath, "apt-get", "-y", "install",
               "--no-install-recommends", "lxc"]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed installing lxc in container %s" %
                             (self.name))

        cmd = ["rm", "-f", "/var/lib/lxc/%s/rootfs/usr/sbin/policy-rc.d" % (self.name)]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError, "Failed removing policy-rc.d in container %s" % (self.name)

    def Execute(self, program="/bin/true"):
        cmd = ["lxc-execute", "-n", self.name, "--", program]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed executing %s in %s" %
                             (program, self.name))

    def Start(self, init="/sbin/init"):
        cmd = ["lxc-start", "-l", "info", "-o", "/tmp/lxcout", "-d", "-n",
               self.name]
        if init != "/sbin/init":
            cmd.append("--")
            cmd.append(init)
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed starting %s in %s" % (init, self.name))

    def ConfirmRunning(self, notrunning=False):
        # ideally lxc-wait would take a timeout and we could do
        # lxc-wait -n test-$release -s RUNNING -t 5.  We don't though.
        time.sleep(2)

        cmd1 = ["lxc-info", "-n", self.name, "-s"]
        sp = subprocess.Popen(cmd1, stderr=subprocess.STDOUT,
                              stdout=subprocess.PIPE)
        output = sp.communicate()[0]
        if not notrunning:
            output.index("RUNNING")  # will raise a ValueError if not found.
        else:
            output.index("STOPPED")  # will raise a ValueError if not found.

    def TrySsh(self):
        True
        # not yet implemented

    # should we be trying a clean shutdown?
    def Stop(self):
        cmd = ["lxc-stop", "-n", self.name]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed stopping %s" % (self.name))

    def apparmor_profile(self, profile):
        # remove any existing lxc.aa_profile lines, and add a new one
        # saying lxc.aa_profile = unconfined.
        configfile = "/var/lib/lxc/%s/config" % (self.name)
        cmd = ["sed", "-i", "/aa_profile/d", configfile]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed to remove existing aa_profile lines")
        newprofile = '''$a\
lxc.aa_profile = unconfined'''
        cmd = ["sed", "-i", newprofile, configfile]
        ret = subprocess.call(cmd, stdout=None, stderr=None)
        if ret != 0:
            raise ValueError("Failed to set unconfined profile")


# a class whose setUp creates a container of a certain suite,
# and tearDown stops and destroys it.
# TODO: add options, like backing store.
class SimpleLxcTests(unittest.TestCase):
    def setUp(self):
        # get lsb release
        dist = platform.dist()
        self.hostrelease = float(dist[1])
        self.arch = platform.machine()
        if self.arch == "x86_64":
            self.arch = "amd64"
        elif self.arch == "i686" or self.arch == "x86":
            self.arch = "i386"
        elif self.arch == "armv7l":
            self.arch = "armel"

        if self.arch == "armel" or self.arch == "armhf":
            self.tests = [["ubuntu", 12.04, "precise"],
                          ["ubuntu", 13.04, "raring"]]
        else:
            self.tests = [["ubuntu", 8.04, "lucid"],
                          ["ubuntu", 12.04, "precise"],
                          ["ubuntu", 13.04, "raring"],
                          #["ubuntu-cloud", 8.04, "lucid"],
                          #["ubuntu-cloud", 12.04, "precise"],
                          ]

    def runTest(self):
        for (t, relnum, r) in self.tests:
            try:
                test = "initialization"
                l = Lxc(template=t, relnum=relnum, release=r)
                test = "destroy preexisting"
                l.Destroy(ignorefail=True)
                test = "create"
                l.Create()
                if t == 'ubuntu':
                    test = "lxc package setup in container"
                    l.PkgSetup()
                    test = "lxc-execute"
                    l.Execute("/bin/true")
                    test = "destroy"
                    l.Destroy()
                    test = "second create"
                    l.Create()
                test = "lxc-start"
                l.Start()
                test = "confirm container is running"
                l.ConfirmRunning()
                test = "ssh into container"
                l.TrySsh()
                test = "stopping the container"
                l.Stop()
                test = "final destroy"
                l.Destroy()
            except ValueError, e:
                self.assertTrue(False, "Failed with release %s at test %s"
                                       "\nError: %s" % (r, test, e))

        return True

        # Create a container and clone it
        try:
            test = "Create original container"
            f = open("lxc:globtest", "w")
            f.close()
            # this also tests a container without '-f config'
            l = Lxc(release="precise", config=None)
            l.Destroy(ignorefail=True)
            l.Create()
            test = "Clone container"
            l2 = Lxc(name="test-clone")
            l2.Destroy(ignorefail=True)
            l2.Clone(clonefrom=l)
            test = "Start cloned container"
            l2.Start()
            test = "Confirm container clone is running"
            l2.ConfirmRunning()
            test = "Stop cloned container"
            l2.Stop()
            test = "Destroy cloned container"
            l2.Destroy()
            test = "Destroy original container"
            l.Destroy()
        except ValueError, e:
            try:
                l.Destroy(ignorefail=True)
                l2.Destroy(ignorefail=True)
            except:
                pass
            self.assertTrue(False, "Failed at container clone test %s\n"
                                   "Error: %s" % (test, e))

        # test container autostart
        try:
            test = "Create autostart container"
            l = Lxc(name="autotest1")
            l.Destroy(ignorefail=True)
            l.Create()
            cmd = ["ln", "-s", "/var/lib/lxc/autotest1/config",
                   "/etc/lxc/auto/autostart1.conf"]
            ret = subprocess.call(cmd, stdout=None, stderr=None)
            if ret != 0:
                return ValueError, "Failed creating autostart symlink"
            test = "stop lxc"
            subprocess.call(["stop", "lxc"], stdout=None, stderr=None)
            test = "start lxc"
            subprocess.call(["start", "lxc"], stdout=None, stderr=None)
            test = "check autostart container is running"
            l.ConfirmRunning()
            test = "stop lxc(2)"
            subprocess.call(["stop", "lxc"], stdout=None, stderr=None)
            l.ConfirmRunning(notrunning=True)
            l.Destroy()
            autolinksurvived = True
            try:
                os.stat("/etc/lxc/auto/autostart1.conf")
            except:
                autolinksurvived = False
            if autolinksurvived:
                raise ValueError("Destroy did not remove auto symlink")
            l.Destroy(ignorefail=True)
        except ValueError, e:
            try:
                l.Destroy(ignorefail=True)
            except:
                pass
            self.assertTrue(False, "Failed at container autostart test %s\n"
                                   "Error: %s" % (test, e))

if __name__ == '__main__':
    # check for uid 0
    uid = os.getuid()
    if uid != 0:
        print >>sys.stderr, "Run this test suite as root"
        sys.exit(1)

    # worry about what to do about other distros later
    dist = platform.dist()
    if dist[0] != "Ubuntu":
        print >>sys.stderr, ("This test suite does not yet support %s" %
                             (dist[0]))
        sys.exit(1)

    # make sure the needed prereqs are installed
    cmd = ["sudo", "apt-get", "-y", "install", "lxc", "make", "gcc"]
    if dist[1] >= 12.10:
        cmd += ["bzr", "python3-lxc"]

    ret = subprocess.call(cmd, stdout=None, stderr=None)
    if ret != 0:
        print "Failed installing prerequisites"
        sys.exit(1)

    # apt-get install -y cgroup-lite libvirt-bin lxc lsb-release make gcc
    if dist[1] < 12.04:
        print >>sys.stderr, "testsuite only supported on precise and later"
        sys.exit(1)

    suite = unittest.TestSuite()

    suite.addTest(SimpleLxcTests())

    rc = unittest.TextTestRunner(verbosity=2).run(suite)

    if not rc.wasSuccessful():
        sys.exit(1)

# set et sw=4 ts=4 ai si sm background=dark
